#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2019 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import string
import random
from chainedci.config import getConfig, init_ini
from chainedci.log import log

'''
Tests library.
'''


def force_ini_values():
    init_ini()
    ini = getConfig('ini')
    ini['log']['level'] = 'DEBUG'
    ini['log']['deprecation_warning'] = True
    ini['log']['exception_or_exit'] = True
    log.setLevel('DEBUG')


def fake_token_selector(url):
    return {'TOKEN_VAR_NAME': 'TOKEN_VALUE'}


def rand_str(stringLength=10):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(stringLength))


def clean_secrets():
    ini = getConfig('ini')
    ini.dl._vault.secrets = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2019 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Init_project module."""

import os
import re
import shutil
import pkg_resources
import git
from jinja2 import Environment, FileSystemLoader
from chainedci.config import getConfig
from chainedci.log import log


def init_project():
    """Init the chainedci structure."""
    ini = getConfig('ini')
    sc_folder = ini['init']['tree']['scenarios']['folder']
    cf_folder = ini['init']['tree']['config']
    create_folders(sc_folder, cf_folder)
    create_all_file(sc_folder)
    create_static_files(sc_folder, cf_folder)


def create_folders(sc_folder, cf_folder):
    """Create chained-ci folders."""
    ini = getConfig('ini')
    folders = [
        f"{sc_folder}",
        f"{sc_folder}/{ini['init']['tree']['scenarios']['group']}",
        f"{sc_folder}/{ini['init']['tree']['scenarios']['definitions']}",
        f"{cf_folder}",
        f"{cf_folder}/{ini['artifacts']['certificates']['src']}",
        f"{cf_folder}/{ini['artifacts']['ssh_creds']['src']}",
        f"{cf_folder}/{ini['artifacts']['pdfidf']['src']}",
        f"{cf_folder}/{ini['artifacts']['ssh_access']['src']}",
        f"{cf_folder}/{ini['artifacts']['static']['src']}",
    ]
    for folder in folders:
        log.info('create folder %s', folder)
        os.makedirs(folder, exist_ok=True)


def create_all_file(sc_folder):
    """Create all.yml configuration."""
    ini = getConfig('ini')
    file_loader = FileSystemLoader(
        pkg_resources.resource_filename('chainedci', 'templates/'))
    env = Environment(autoescape=True, loader=file_loader)
    ci_template = env.get_template('all.yml.tpl')
    remote = git.Repo('.').remotes['origin'].url
    match = re.search(r'.*@(.*):(.*)\.git', remote)
    # SET_PROJECT_TRIGGER_IF_NEEDED is a placeholder
    output = ci_template.render( # nosec B106
        current_chained_ci_version='master',
        project_long_name=match.group(2).replace('/', '_'),
        gitlab_server=match.group(1),
        config_step=ini['specific_steps']['config'],
        config_path=ini['init']['tree']['config'],
        trigger_step=ini['specific_steps']['trigger_myself'],
        trigger_token='SET_PROJECT_TRIGGER_IF_NEEDED',
        example_project='/'.join(match.group(2).split('/')[:-1])+'/project_A'
    )
    filename = (f"{sc_folder}/{ini['init']['tree']['scenarios']['group']}"
                "/all.yml")
    with open(filename, 'w', encoding="utf-8") as all_file:
        all_file.write(output)
        log.info("Write projects file to '%s'", filename)


def create_static_files(sc_folder, cf_folder):
    """Create static init files."""
    ini = getConfig('ini')
    root = pkg_resources.resource_filename('chainedci', 'static/')
    shutil.copyfile(f"{root}/projectA.yml",
                    (f"{sc_folder}/"
                     f"{ini['init']['tree']['scenarios']['definitions']}"
                     "/projectA.yml"))
    shutil.copyfile(f"{root}/inventory",
                    (f"{sc_folder}"
                     "/inventory"))
    shutil.copyfile(f"{root}/projectA_config1.yml",
                    (f"{cf_folder}/{ini['artifacts']['pdfidf']['src']}"
                     "/projectA_config1.yml"))
    shutil.copyfile(f"{root}/projectA_config2.yml",
                    (f"{cf_folder}/{ini['artifacts']['pdfidf']['src']}"
                     "/projectA_config2.yml"))



#!/usr/bin/env python3

"""Chained CI Artifact source."""

from shutil import copy, copytree, rmtree, unpack_archive, ReadError
from tempfile import mkdtemp
from os import makedirs
from os.path import isdir, exists
import binascii
import requests
from schema import Schema, Or, Optional
from chainedci.config import getConfig
from chainedci.tools import token_selector, get_env_or_raise, raise_ex
from chainedci.log import log
from chainedci.http_adapter import Http
from chainedci.parser import JinjaParser


class ArtifactSrc(dict):
    """Artifact Source object."""

    def __init__(self, src):
        """
        Artifact src init.

        :param src: The source of the artifact
        :type src: dict
        :param artifact_folder: The destination folder
        :type artifact_folder: string
        """
        super().__init__()
        self.uuid = binascii.crc32(str(src).encode('ascii'))
        self._folder = None
        self.update(self._validate(src))
        log.debug("ArtifactSrc[%s]: init type '%s' with source '%s'",
                  self.uuid, self['type'], self['source'])

    def load(self, artifact_folder):
        """
        Pull the ArtifactSrc locally.

        :param artifact_folder: The destination folder
        :type artifact_folder: string
        """
        # First lookup and replace variables in source
        parser = JinjaParser()
        self['source'] = parser.parse(self['source'])
        # Start loading
        log.debug("ArtifactSrc[%s]: load %s", self.uuid, self['source'])
        self._folder = artifact_folder
        if self._folder is None:
            raise ValueError("Artifact folder was never set")
        if self['type'] == 'pipeline':
            self._load_from_pipeline(self._folder)
        elif self['type'] == 'remote':
            self._load_from_remote()
        elif self['type'] == 'local_archive':
            self.unpack(self._folder, remote=False)
        elif self['type'] == 'remote_archive':
            self.unpack(self._folder)
        else:
            self._load_from_local()

    def _validate(self, src):
        """Validate ArtifactSrc Schema."""
        log.debug("ArtifactSrc[%s]: validation", self.uuid)
        schema = Schema(Or({'type': 'local',
                            'source': str,
                            'destination': str},
                           {'type': 'remote',
                            'source': str,
                            'destination': str},
                           {'type': 'local_archive',
                            'source': str},
                           {'type': 'remote_archive',
                            'source': str},
                           {'type': 'pipeline',
                            'source': str,
                            Optional('limit_to'): [dict],
                            Optional('artifact_in_pipeline'): bool}))
        return schema.validate(src)

    def unpack(self, dest='./', remote=True):
        """
        Download remote archive and unpack in a local folder.

        :param dest: The destination folder
        :type dest: string
        """
        log.debug("ArtifactSrc[%s]: get file to unpack %s to %s", self.uuid,
                  self['source'], dest)
        tmp_folder = mkdtemp()
        if remote:
            self._load_from_remote(tmp_folder + "/archive.zip")
        else:
            self._load_from_local(tmp_folder + "/archive.zip")

        log.debug("ArtifactSrc[%s]: unpack %s to %s", self.uuid,
                  tmp_folder + "/archive.zip", dest)
        try:
            unpack_archive(tmp_folder + "/archive.zip", dest)
        except ReadError:
            raise_ex(ValueError, f"{self['source']} is not an archive file")

        rmtree(tmp_folder)

    def _load_from_local(self, destination=None, source=None):
        """Pull the ArtifactSrc locally."""
        src = source or self['source']
        dest = destination or f"{self._folder}/{self['destination']}"
        log.debug("ArtifactSrc[%s]: load '%s' from local to '%s'", self.uuid,
                  src, dest)
        if not isdir(self._folder):
            raise OSError(f'Folder {self._folder} does not exists')
        # create parent directory
        makedirs('/'.join(dest.split('/')[:-1]), exist_ok=True)
        if isdir(src):
            copytree(src, dest, dirs_exist_ok=True)
        else:
            copy(src, dest)

    def _download_file(self, dest):
        """Download the file from an url."""
        log.debug("ArtifactSrc[%s]: download file '%s' to '%s'", self.uuid,
                  self['source'], dest)
        headers = token_selector(self['source'])
        with Http().session.get(self['source'], headers=headers) as response, \
                open(dest, 'wb') as out_file:
            # pylint: disable=no-member
            if response.status_code != requests.codes.ok:
                raise_ex(requests.exceptions.HTTPError,
                         f"ArtifactSrc: error calling '{self['source']}':"
                         f" '{ response.status_code }'")
            out_file.write(response.content)
            log.debug("ArtifactSrc[%s]: download done", self.uuid)

    def _load_from_remote(self, dest=None):
        """Load a remote artifact."""
        log.debug("ArtifactSrc[%s]: load remote '%s'", self.uuid,
                  self['source'])
        if dest is None:
            dest = f"{self._folder}/{self['destination']}"
        self._download_file(dest)

    def _load_from_pipeline(self, artifact_folder=None):
        """Pull the pipeline artifact."""
        log.debug("ArtifactSrc[%s]: load from pipeline '%s' to '%s'",
                  self.uuid, self['source'], artifact_folder)
        if self.get('artifact_in_pipeline', True):
            self._load_from_my_pipeline(artifact_folder)
        else:
            self._load_from_another_pipeline(artifact_folder)

    def _api_get(self, url):
        """Call an API."""
        log.debug("ArtifactSrc[%s]: call api '%s'", self.uuid, url)
        headers = token_selector(url)
        with Http().session.get(url, headers=headers) as response:
            # pylint: disable=no-member
            if response.status_code != requests.codes.ok:
                raise_ex(requests.exceptions.HTTPError,
                         f"ArtifactSrc[{self.uuid}]: error calling '{url}':"
                         f" '{ response.status_code }'")
            return response.json()

    def _load_from_my_pipeline(self, destination):
        """Load files from a previous job in this pipeline."""
        run = getConfig('run')
        pipeline_data = self._get_pipeline_successful_jobs()
        # filter steps to get the last sucessful step id
        target_step = f"{ self['source'] }:{ run['name'] }"
        last_success = list(filter(lambda d: d['name'] in target_step,
                                   pipeline_data))
        if not last_success:
            raise_ex(ValueError, f"ArtifactSrc[{self.uuid}]"
                                 f"No job found with name '{target_step}'")
        last_success_id = last_success[-1]['id']
        if 'limit_to' in self:
            tmp_folder = mkdtemp()
            self._download_job_artifact(last_success_id, tmp_folder)
            self._limit_to(tmp_folder, destination)
        else:
            self._download_job_artifact(last_success_id, destination)

    def _get_pipeline_successful_jobs(self, pipeline_id=None, project_id=None):
        """Get the list of a pipeline sucessful jobs."""
        # Get ids from env vars
        log.debug("ArtifactSrc[%s]: Search for pipeline '%s' successful jobs",
                  self.uuid, pipeline_id)
        config = getConfig('chainedci')
        api = config['defaults']['gitlab']['api_url']
        project_id = project_id or get_env_or_raise('CI_PROJECT_ID')
        pipeline_id = pipeline_id or get_env_or_raise('CI_PIPELINE_ID')
        # set request url
        url = (f"{ api }/projects/{ project_id }/"
               f"pipelines/{ pipeline_id }/jobs?scope[]=success")
        return self._api_get(url)

    def _download_job_artifact(self, job_id, destination, project_id=None):
        """Download an artifact of a job."""
        log.debug("ArtifactSrc[%s]: Download job '%s'", self.uuid,
                  job_id)
        config = getConfig('chainedci')
        api = config['defaults']['gitlab']['api_url']
        project_id = project_id or get_env_or_raise('CI_PROJECT_ID')
        self['source'] = (f"{ api }/projects/{ project_id }/"
                          f"jobs/{ job_id }/artifacts")
        self.unpack(destination)

    def _limit_to(self, tmp_folder, destination):
        """Filter the files of an archive."""
        log.debug("ArtifactSrc[%s]: Filter files. Move from '%s' to '%s'",
                  self.uuid, tmp_folder, destination)
        for file_filter in self['limit_to']:
            file_src = next(iter(file_filter))
            if not exists(f'{tmp_folder}/{file_src}'):
                raise_ex(ValueError, f"The file '{file_src}' does not exists "
                         "in this artifact.")
            self._load_from_local(
                source=f'{tmp_folder}/{file_src}',
                destination=f'{destination}/{file_filter[file_src]}')

    def _load_from_another_pipeline(self, artifact_folder):
        return artifact_folder



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2019 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Inventory module."""


from os import path
from ansible.inventory.manager import InventoryManager
from ansible.module_utils._text import to_bytes
from ansible.parsing.dataloader import DataLoader
from ansible.parsing.vault import VaultSecret
from chainedci.config import Config, getConfig
from chainedci.log import log, add_tokens_to_log_filter


class Inventory():
    """Inventory object."""

    def __init__(self, inventory_file):
        """Init Inventory."""
        self.config = None
        self.inventory_file = inventory_file
        self.get_inventory_manager()
        self.scenario_name = None
        self.job = None

    def load(self, scenario_name, job_name):
        """Load inventory files and hide tokens."""
        self.scenario_name = scenario_name
        self.job = job_name
        self.load_config(scenario_name, job_name)
        run = getConfig('run')
        add_tokens_to_log_filter(self.config, [run.get('key')])
        config = getConfig('chainedci')
        config.update_key(self.config)

    def get_inventory_manager(self, key_id='default'):
        """Load ansible inventory manager."""
        # all needs loader
        loader = DataLoader()
        run = getConfig('run')

        if run.get('key'):
            secret = [(key_id,
                       VaultSecret(to_bytes(run['key'], encoding='utf-8')))]
            loader.set_vault_secrets(secret)

        self.inventory = InventoryManager(loader=loader,
                                          sources=[self.inventory_file])

    def load_config(self, scenario_name, job_name):
        """Load all config files depending on inventory file."""
        self.config = Config(load_env=False, init_scenarios=True)
        self.load_vault_key()

        # load configs
        host = self.inventory.hosts[scenario_name]
        inv_folder = host.vars['inventory_dir']
        self.load_groups(inv_folder, host)
        self.load_host_file(inv_folder, host)

        # set some run param in run config
        run = getConfig('run')
        run['name'] = scenario_name
        run['job'] = job_name
        self.config.simplify()

    def load_groups(self, inv_folder, host):
        """Load group files."""
        groups_to_load = host.get_groups()
        all_file = f"{inv_folder}/group_vars/all.yml"
        log.info('Load global group file "%s"', all_file)
        self.config.load_file(all_file, 'defaults')
        for grp in groups_to_load:
            if grp.name not in ['ungrouped', 'all']:
                self.load_group(inv_folder, grp.name)

    def load_group(self, inv_folder, group_name):
        """Load a group file."""
        file = f"{inv_folder}/group_vars/{group_name}.yml"
        if path.exists(file):
            log.info('Load group file "%s"', file)
            self.config.merge(file, 'defaults')
        else:
            log.info('No group file "%s"', file)

    def load_host_file(self, inv_folder, host):
        """Load host file."""
        host_file = f"{inv_folder}/host_vars/{host.name}.yml"
        log.info('Load scenario file "%s"', host_file)
        self.config.load_file(host_file, f'scenarios.{host.name}')

    def load_vault_key(self):
        """Load vault key if needed."""
        run = getConfig('run')
        if run.get('key'):
            log.debug('load vault key')
            self.config.set_vault_key(run['key'])
        else:
            self.config.clean_vault_keys()

    @property
    def scenarios(self):
        """List hosts extracted from inventory."""
        return self.inventory.hosts



#!/usr/bin/env python3

"""CLI functions for running a job or generate a CI config."""

from os import makedirs
import pkg_resources
from jinja2 import Environment, FileSystemLoader
from chainedci.config import getConfig
from chainedci.inventory import Inventory
from chainedci.init_project import init_project
from chainedci.log import log
from chainedci.options import Options
from chainedci.scenario import Scenario
from chainedci.tools import raise_ex
from chainedci.version import __version__


def main():
    """Read sys opts and run cli commands."""
    opts = Options()
    inv = Inventory(opts.opts.inventory)
    if opts.opts.mode == 'run':
        log.info("Start running the scenario job")
        run(opts, inv)
    elif opts.opts.mode == 'generate':
        log.info("Start generating the ci file")
        generate(inv, opts.opts.scenario_name)
    elif opts.opts.mode == 'init':
        log.info("Generate the chained-ci structure")
        init_project()
    else:
        raise_ex(ValueError, f'Unknown command "{opts.opts.mode}"')


def _generate_scenario(env, defaults, scenario_name,
                       scenario_config):
    """
    Generate scenarios into gitlab-ci.yml files.

    :param env: Jinja environment
    :type env: jinja environment
    :param defaults: Default configuration from all.yml
    :type defaults: Config
    :param scenario_name: The scenario name
    :type scenario_name: string
    :param scenarios_config: the scenario config
    :type scenarios_config: Config object
    """
    ini = getConfig('ini')
    sc_template = env.get_template('scenario.yml.tpl')
    stages = []
    for step_config in scenario_config['scenario_steps'].values():
        stage = (
            step_config['stage'] if (
                'stage' in step_config and
                ini['generator']['ci_dynamic_stages'] and
                ini['generator']['mode'] == 'dynamic')
            else defaults['gitlab']['git_projects'][step_config[
                'project']]['stage'])
        step_config['stage'] = stage
        if stage not in stages:
            stages.append(stage)
    output = sc_template.render(
        ci_mode=ini['generator']['mode'],
        ci_vault_input=ini['generator']['ci_vault_input'],
        ci_vault_file_var=ini['encryption']['key_env_name'],
        stages=stages,
        runner=defaults['runner'],
        scenario_name=scenario_name,
        scenario=scenario_config,
        git_projects=defaults['gitlab']['git_projects'],
        header=(f"Scenario {scenario_name} generated by chained-ci-py")
    )
    filename = (ini['generator']['ci_dynamic_file']
                if ini['generator']['mode'] == 'dynamic' else
                f"{ini['generator']['ci_included_folder']}/"
                f"{ini['generator']['ci_included_prefix']}"
                f"{scenario_name}.yml")
    with open(filename, 'w', encoding="utf-8") as scenario_file:
        scenario_file.write(output)
        log.info("Write scenario %s to '%s'", scenario_name, filename)


def _generate_main(env, defaults, scenarios_list, scenarios_config):
    """
    Generate scenarios into gitlab-ci.yml files.

    :param env: Jinja environment
    :type env: jinja environment
    :param defaults: Default configuration from all.yml
    :type defaults: Config
    :param scenarios_list: The scenario names
    :type scenarios_list: list of string
    :param scenarios_config: the scenario config
    :type scenarios_config: Config object
    """
    ini = getConfig('ini')
    ci_template = env.get_template('main.yml.tpl')
    output = ci_template.render(
        ci_mode=ini['generator']['mode'],
        yaml_image=ini['generator']['images']['yaml_checking'],
        ci_vault_input=ini['generator']['ci_vault_input'],
        ci_vault_file_var=ini['encryption']['key_env_name'],
        stages=defaults['stages'],
        runner=defaults['runner'],
        disable_pages=defaults['disable_pages'],
        scenarios=scenarios_list,
        scenarios_cfg=scenarios_config,
        ci_include_prefix=(f"{ini['generator']['ci_included_folder']}/"
                           f"{ini['generator']['ci_included_prefix']}"),
        header=("Generated by chained-ci-py")
    )
    with open(f"{ini['generator']['gitlabci_main']}", 'w',
              encoding="utf-8") as ci_file:
        ci_file.write(output)
        log.info("Write CI main file to '%s'",
                 f"{ini['generator']['gitlabci_main']}")


def generate(inv, scenario_name=None):
    """
    Generate scenarios into gitlab-ci.yml files.

    :param inv: Inventory generaterd from opts
    :type timeout: Inventory class
    :param scenario_name: The scenarion name from opts if 'dynamic' mode
    :type scenario_name: string
    """
    file_loader = FileSystemLoader(
        pkg_resources.resource_filename('chainedci', 'templates/'))
    env = Environment(autoescape=True, loader=file_loader)
    ini = getConfig('ini')
    if ini['generator']['mode'] == 'dynamic' and scenario_name:
        inv.load_config(scenario_name, None)
        _generate_scenario(env, inv.config['defaults'],
                           scenario_name,
                           inv.config['scenarios'][scenario_name])
    elif ini['generator']['mode'] == 'dynamic' and scenario_name is None:
        inv.load_config(list(inv.scenarios.keys())[0], None)
        _generate_main(env, inv.config['defaults'], [], {})
    else:
        scenarios = {}
        for scenario in inv.scenarios:
            inv.load_config(scenario, None)
            scenarios[scenario] = inv.config['scenarios'][scenario]
        if ini['generator']['mode'] == 'splitted':
            makedirs(ini['generator']['ci_included_folder'], exist_ok=True)
            for scenario in inv.scenarios:
                _generate_scenario(env, inv.config['defaults'],
                                   scenario, scenarios[scenario])
        _generate_main(env, inv.config['defaults'],
                       inv.scenarios, scenarios)


def run(opts, inv):
    """
    Run scenario step.

    :param opts: Options from command line
    :type timeout: Option class
    :param inv: Inventory generaterd from opts
    :type timeout: Inventory class
    """
    inv.load(opts.opts.scenario_name, opts.opts.job)
    scenario = Scenario(opts.opts.scenario_name)
    scenario.run(opts.opts.job)
